#include "timestep.h"
#include <vector>
#include <amp.h>
#include "write_file.h"

using namespace std;
using namespace concurrency;

int drive_lid(const struct st_parameters param, concurrency::array<struct grid_cell1, 1> &amp_node1, concurrency::array<struct grid_cell2, 1> &amp_node2, concurrency::array<grid_cell3, 1> &amp_node3, const concurrency::array<int, 1> &amp_obst) {
	// Modify the first column of the grid
    parallel_for_each(
        // Define the compute domain, which is the set of threads that are created.
		amp_node1.extent,
        // Define the code to run on each thread on the accelerator.
		[&amp_node1, &amp_node2, &amp_node3, &amp_obst, param](index<1> idx) restrict(amp)
        {
		float w0,w1,w2;

		// Calculate weighting factors
		w0 = 4.0f / 9.0f;
		w1 = 1.0f / 9.0f;
		w2 = 1.0f / 36.0f;
	
		float c_sq = 1.0f/3.0f;	// Square of speed of sound
		float u[9];
		float u_x = param.accel;

		// Make sure the cell is not blocked, and 
		// that we do not send a negative density.
		if(idx[0] > (param.ny*param.nx-param.nx-1) && amp_obst[idx] == 1) {
				// Directional velocity components;
				float u_sq = u_x*u_x + 0.0f*0.0f;
				u[1] =  u_x;		// East
				u[2] =        0.0f;	// North
				u[3] = -u_x;		// West
				u[4] =      - 0.0f;	// South
				u[5] =  u_x + 0.0f;	// North-East
				u[6] = -u_x + 0.0f;	// North-West
				u[7] = -u_x - 0.0f;	// South-West
				u[8] =  u_x - 0.0f;	// South-East

				// Equalibrium densities
				// Zero velocity density: weight w0
				amp_node2[idx].c = w0 * param.density * (1.0f - u_sq / (2.0f * c_sq));
				// Axis speeds: weight w1
				amp_node2[idx].e = w1 * param.density * (1.0f + u[1] / c_sq
									+ (u[1] * u[1]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node3[idx].n = w1 * param.density * (1.0f + u[2] / c_sq
									+ (u[2] * u[2]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node2[idx].w = w1 * param.density * (1.0f + u[3] / c_sq
									+ (u[3] * u[3]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node1[idx].s = w1 * param.density * (1.0f + u[4] / c_sq
									+ (u[4] * u[4]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				// Diagonal speeds: weight w2
				amp_node3[idx].ne = w2 * param.density * (1.0f + u[5] / c_sq
									+ (u[5] * u[5]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node3[idx].nw = w2 * param.density * (1.0f + u[6] / c_sq
									+ (u[6] * u[6]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node1[idx].sw = w2 * param.density * (1.0f + u[7] / c_sq
									+ (u[7] * u[7]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node1[idx].se = w2 * param.density * (1.0f + u[8] / c_sq
									+ (u[8] * u[8]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
			}
	}
	);
	return 1;
}

int inout3(const struct st_parameters param, concurrency::array<struct grid_cell1, 1> &amp_node1, concurrency::array<struct grid_cell2, 1> &amp_node2, concurrency::array<grid_cell3, 1> &amp_node3, const concurrency::array<int, 1> &amp_obst) {
	// Modify the first column of the grid
    parallel_for_each(
        // Define the compute domain, which is the set of threads that are created.
		amp_node1.extent,
        // Define the code to run on each thread on the accelerator.
		[&amp_node1, &amp_node2, &amp_node3, &amp_obst, param](index<1> idx) restrict(amp)
        {
		float w0,w1,w2;

		// Calculate weighting factors
		w0 = 4.0f / 9.0f;
		w1 = 1.0f / 9.0f;
		w2 = 1.0f / 36.0f;
	
		float c_sq = 1.0f/3.0f;	// Square of speed of sound
		float u[9];
		float u_x = param.accel;

		// Make sure the cell is not blocked, and 
		// that we do not send a negative density.
		if(float((idx[0]+1) % param.nx) == 0.00f && amp_obst[idx] == 0) {
				// Directional velocity components;
				float u_sq = u_x*u_x + 0.0f*0.0f;
				u[1] =  u_x;		// East
				u[2] =        0.0f;	// North
				u[3] = -u_x;		// West
				u[4] =      - 0.0f;	// South
				u[5] =  u_x + 0.0f;	// North-East
				u[6] = -u_x + 0.0f;	// North-West
				u[7] = -u_x - 0.0f;	// South-West
				u[8] =  u_x - 0.0f;	// South-East

				// Equalibrium densities
				// Zero velocity density: weight w0
				amp_node2[idx].c = w0 * param.density * (1.0f - u_sq / (2.0f * c_sq));
				// Axis speeds: weight w1
				amp_node2[idx].e = w1 * param.density * (1.0f + u[1] / c_sq
									+ (u[1] * u[1]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node3[idx].n = w1 * param.density * (1.0f + u[2] / c_sq
									+ (u[2] * u[2]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node2[idx].w = w1 * param.density * (1.0f + u[3] / c_sq
									+ (u[3] * u[3]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node1[idx].s = w1 * param.density * (1.0f + u[4] / c_sq
									+ (u[4] * u[4]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				// Diagonal speeds: weight w2
				amp_node3[idx].ne = w2 * param.density * (1.0f + u[5] / c_sq
									+ (u[5] * u[5]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node3[idx].nw = w2 * param.density * (1.0f + u[6] / c_sq
									+ (u[6] * u[6]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node1[idx].sw = w2 * param.density * (1.0f + u[7] / c_sq
									+ (u[7] * u[7]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
				amp_node1[idx].se = w2 * param.density * (1.0f + u[8] / c_sq
									+ (u[8] * u[8]) / (2.0f * c_sq * c_sq)
									- u_sq / (2.0f * c_sq));
			}
	}
	);
	return 1;
}

int twolattice(const struct st_parameters param, const concurrency::array<struct grid_cell1, 1> &amp_node1, const concurrency::array<struct grid_cell2, 1> &amp_node2, const concurrency::array<grid_cell3, 1> &amp_node3, concurrency::array<struct grid_cell1, 1> &amp_tmp1, concurrency::array<struct grid_cell2, 1> &amp_tmp2, concurrency::array<grid_cell3, 1> &amp_tmp3, const concurrency::array<int, 1> &amp_obst) {
    parallel_for_each(
        // Define the compute domain, which is the set of threads that are created.
		amp_tmp1.extent,
        // Define the code to run on each thread on the accelerator.
		[&amp_node1, &amp_node2, &amp_node3, &amp_tmp1, &amp_tmp2, &amp_tmp3, &amp_obst, param](index<1> idx) restrict(amp)
        {
			float c_sq = 1.0f/3.0f;		// Square speed of sound
			float w0 = 4.0f/9.0f;		// Weighting factors
			float w1 = 1.0f/9.0f;
			float w2 = 1.0f/36.0f;

			// Propagate
			int n = ( idx[0] < (param.ny*param.nx-param.nx) ) ? idx[0]+param.nx : (param.nx+idx[0] - param.nx*param.ny);
			int s = ( idx[0] > param.nx-1 ) ? idx[0]-param.nx : (param.nx*(param.ny-1) + idx[0]);
			int e_x = ( float((idx[0]+1) % param.nx) != 0.00f ) ? 1 : -param.nx+1; 
			int w_x = ( float((idx[0]) % param.nx) != 0.00f ) ? -1 : param.nx-1; 

			int e = idx[0]+e_x;
			int w = idx[0]+w_x;
			int nw = n+w_x;
			int ne = n+e_x;
			int sw = s+w_x;
			int se = s+e_x;

			float V0, V1, V2, V3, V4, V5, V6,V7,V8;
			V0 = amp_node2(idx).c;
			V1 = amp_node2(w).e;
			V2 = amp_node3(s).n;
			V3 = amp_node2(e).w;
			V4 = amp_node1(n).s;
			V5 = amp_node3(sw).ne;
			V6 = amp_node3(se).nw;
			V7 = amp_node1(ne).sw;
			V8 = amp_node1(nw).se;

			// Bounceback
			if (amp_obst[idx] == 1) {
				// If node is blocked, reverse value
				amp_tmp2[idx].c = V0;
				amp_tmp2[idx].e = V3;
				amp_tmp3[idx].n = V4;
				amp_tmp2[idx].w = V1;
				amp_tmp1[idx].s = V2;
				amp_tmp3[idx].ne = V7;
				amp_tmp3[idx].nw = V8;
				amp_tmp1[idx].sw = V5;
				amp_tmp1[idx].se = V6;
			}

			// Collision			
			if (amp_obst(idx) == 0) {
				float u_x,u_y;				// Avrage velocities in x and y direction
				float u[9];				// Directional velocities
				float d_equ[9];			// Equalibrium densities
				float u_sq;				// Squared velocity
				float local_density;		// Sum of densities in a particular node

				// Calculate local density
				local_density = 0.0;
				local_density += V1;
				local_density += V2;
				local_density += V3;
				local_density += V4;
				local_density += V5;
				local_density += V6;
				local_density += V7;
				local_density += V8;
				local_density += V0;

				// Calculate x velocity component
				u_x = (V1+V5+V8-(V3+V6+V7))/local_density;
				// Calculate y velocity component
				u_y = (V2+V5+V6-(V4+V7+V8))/local_density;
				// Velocity squared
				u_sq = u_x*u_x +u_y*u_y;
				
				// Directional velocity components;
				u[1] =  u_x;		// East
				u[2] =        u_y;	// North
				u[3] = -u_x;		// West
				u[4] =      - u_y;	// South
				u[5] =  u_x + u_y;	// North-East
				u[6] = -u_x + u_y;	// North-West
				u[7] = -u_x - u_y;	// South-West
				u[8] =  u_x - u_y;	// South-East

				// Equalibrium densities
				// Zero velocity density: weight w0
				d_equ[0] = w0 * local_density * (1.0f - u_sq / (2.0f * c_sq));
				// Axis speeds: weight w1
				d_equ[1] = w1 * local_density * (1.0f + u[1] / c_sq
								 + (u[1] * u[1]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));
				d_equ[2] = w1 * local_density * (1.0f + u[2] / c_sq
								 + (u[2] * u[2]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));
				d_equ[3] = w1 * local_density * (1.0f + u[3] / c_sq
								 + (u[3] * u[3]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));
				d_equ[4] = w1 * local_density * (1.0f + u[4] / c_sq
								 + (u[4] * u[4]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));
				// Diagonal speeds: weight w2
				d_equ[5] = w2 * local_density * (1.0f + u[5] / c_sq
								 + (u[5] * u[5]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));
				d_equ[6] = w2 * local_density * (1.0f + u[6] / c_sq
								 + (u[6] * u[6]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));
				d_equ[7] = w2 * local_density * (1.0f + u[7] / c_sq
								 + (u[7] * u[7]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));
				d_equ[8] = w2 * local_density * (1.0f + u[8] / c_sq
								 + (u[8] * u[8]) / (2.0f * c_sq * c_sq)
								 - u_sq / (2.0f * c_sq));

				amp_tmp2[idx].c = (V0 + param.omega * (d_equ[0] - V0));
				amp_tmp2[idx].e = (V1 + param.omega * (d_equ[1] - V1));
				amp_tmp3[idx].n = (V2 + param.omega * (d_equ[2] - V2));
				amp_tmp2[idx].w = (V3 + param.omega * (d_equ[3] - V3));
				amp_tmp1[idx].s = (V4 + param.omega * (d_equ[4] - V4));
				amp_tmp3[idx].ne = (V5 + param.omega * (d_equ[5] - V5));
				amp_tmp3[idx].nw = (V6 + param.omega * (d_equ[6] - V6));
				amp_tmp1[idx].sw = (V7 + param.omega * (d_equ[7] - V7));
				amp_tmp1[idx].se = (V8 + param.omega * (d_equ[8] - V8));
			}
		}
    );

	return 1;
}

void generate_grid3(struct st_parameters param, vector<struct grid_cell1> &node1, vector<struct grid_cell2> &node2, vector<struct grid_cell3> &node3, vector<struct grid_cell1> &tmp_node1, vector<struct grid_cell2> &tmp_node2, vector<struct grid_cell3> &tmp_node3) {
    int x,y;
    int i; // Grid index (y*RowSize+x)
    float w0,w1,w2;

    // Main Grid
    int sizeToAlloc = param.nx*param.ny;
	node1.resize(sizeToAlloc);
	node2.resize(sizeToAlloc);
	node3.resize(sizeToAlloc);

	// Temp grid, used to save scratch values
	tmp_node1.resize(sizeToAlloc);
	tmp_node2.resize(sizeToAlloc);
	tmp_node3.resize(sizeToAlloc);

	// Initialise densities
	w0 = param.density * 4.0/9.0;
	w1 = param.density * 1.0/9.0;
	w2 = param.density * 1.0/36.0;

    for (y=0;y < param.ny;y++) {
    	for(x=0;x<param.nx;x++) {
    		i = y*param.nx + x;
			node2[i].c = w0;

			node2[i].e = w1;
			node3[i].n = w1;
			node2[i].w = w1;
			node1[i].s = w1;

			node3[i].nw = w2;
			node3[i].ne = w2;
			node1[i].sw = w2;
			node1[i].se = w2;
    	}
    }
}

float av_velocity3(const struct st_parameters param, std::vector<struct grid_cell1> &node1, std::vector<struct grid_cell2> &node2, std::vector<struct grid_cell3> &node3, const std::vector<int> &obstacle)
{
	int x,y;
	int i = 0;
    float local_density;  /* total density in cell */
    float tot_u_x;        /* accumulated x-components of velocity */
	int tot_cells = 0;

	/* initialise */
    tot_u_x = 0.0;

	x = (int)((float)param.nx / 2.0);

	for(y=0;y<param.ny;y++) {
		i = y*param.nx + x;
		if (obstacle[i] != 1) {
			local_density = 0.0;
			local_density += node3[i].n;
			local_density += node2[i].e;
			local_density += node1[i].s;
			local_density += node2[i].w;
			local_density += node3[i].ne;
			local_density += node1[i].se;
			local_density += node1[i].sw;
			local_density += node3[i].nw;
			local_density += node2[i].c;

			tot_u_x += (node2[i].e + node3[i].ne + node1[i].se
			- (node2[i].w + node3[i].nw + node1[i].sw)) / local_density;
			 
			++tot_cells;
		}
	}
  return tot_u_x / (float)tot_cells;
}

float total_density3(const struct st_parameters param, std::vector<struct grid_cell1> &node1, std::vector<struct grid_cell2> &node2, std::vector<struct grid_cell3> &node3)
{
  int i, ii,jj;        /* generic counters */
  float total = 0.0;  /* accumulator */
  float x = 0.0;

  for(ii=0;ii<param.ny;ii++) {
    for(jj=0;jj<param.nx;jj++) {
		i = ii*param.nx + jj;
		total += node3[i].n;
		total += node2[i].e;
		total += node1[i].s;
		total += node2[i].w;
		total += node3[i].ne;
		total += node1[i].se;
		total += node1[i].sw;
		total += node3[i].nw;
		total += node2[i].c;
    }
  }

  return total;
}

double calc_reynolds3(const struct st_parameters param, std::vector<struct grid_cell1> &node1, std::vector<struct grid_cell2> &node2, std::vector<struct grid_cell3> &node3, const std::vector<int> &obstacle)
{
  const double viscosity = 1.0 / 6.0 * (2.0 / param.omega - 1.0);
  
  return av_velocity3(param, node1, node2, node3, obstacle) * param.reynolds_dim / viscosity;
}

int timestep3(const struct st_parameters param, int time_save, int out_save) {
	vector<int> obstacle;
	vector<struct grid_cell1> node1;
	vector<struct grid_cell2> node2;
	vector<struct grid_cell3> node3;

	vector<struct grid_cell1> tmp_node1;
	vector<struct grid_cell2> tmp_node2;
	vector<struct grid_cell3> tmp_node3;

	generate_grid3(param, node1, node2, node3, tmp_node1, tmp_node2, tmp_node3);
	read_obstacles2(param, obstacle);

	int n_nodes = param.nx*param.ny;
	accelerator device = accelerator();
	accelerator_view av = device.create_view(queuing_mode_immediate);

	array<grid_cell1, 1> amp_node1(n_nodes, node1.begin(), node1.end(), av);
	array<grid_cell2, 1> amp_node2(n_nodes, node2.begin(), node2.end(), av);
	array<grid_cell3, 1> amp_node3(n_nodes, node3.begin(), node3.end(), av);

	array<grid_cell1, 1> amp_tmp1(n_nodes, tmp_node1.begin(), tmp_node1.end(), av);
	array<grid_cell2, 1> amp_tmp2(n_nodes, tmp_node2.begin(), tmp_node2.end(), av);
	array<grid_cell3, 1> amp_tmp3(n_nodes, tmp_node3.begin(), tmp_node3.end(), av);

	array<int, 1> amp_obst(n_nodes, obstacle.begin(), obstacle.end(), av);

	cout << endl << "C++ AMP Optimized" << endl;
	clock_t start;
	float diff;
	float mlups;
	double time_now;
	int step_now;
	int p;

	time_now = clock();
	step_now = 0;
	start = clock();
	unsigned long long t;

	for(int i=0;i<param.t_max;i++) {
		if (i%time_save==0)
        {
			float time = (clock())/((float)CLOCKS_PER_SEC);
			t = (unsigned long long)param.nx*(unsigned long long)param.ny*(unsigned long long)i;
			mlups = (float)t / (time*1000000);
			p = 100*i/param.t_max;
			cout << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
			cout << p << "%\t Avg. MLUP/s " <<mlups; 
		}
		
		if (i%2==0) {
			if (out_save != 0 && i%out_save==0)
			{
				int sa = i/out_save;
				concurrency::copy(amp_node1, node1.begin());
				concurrency::copy(amp_node2, node2.begin());
				concurrency::copy(amp_node3, node3.begin());
				write_file3(param, node1, node2, node3, obstacle, sa, "ampo");
			}
			//drive_lid(param,amp_node1, amp_node2, amp_node3, amp_obst);
			inout3(param,amp_node1, amp_node2, amp_node3, amp_obst);
			twolattice(param, amp_node1, amp_node2, amp_node3, amp_tmp1, amp_tmp2, amp_tmp3, amp_obst);
		} else {
			if (out_save != 0 && i%out_save==0)
			{
				int sa = i/out_save;
				concurrency::copy(amp_tmp1, node1.begin());
				concurrency::copy(amp_tmp2, node2.begin());
				concurrency::copy(amp_tmp3, node3.begin());
				write_file3(param, node1, node2, node3, obstacle, sa, "ampo");
			}
			//drive_lid(param,amp_tmp1, amp_tmp2, amp_tmp3, amp_obst);
			inout3(param,amp_tmp1, amp_tmp2, amp_tmp3, amp_obst);
			twolattice(param, amp_tmp1, amp_tmp2, amp_tmp3, amp_node1, amp_node2, amp_node3, amp_obst);
		}
	}

	concurrency::copy(amp_node1, node1.begin());
	concurrency::copy(amp_node2, node2.begin());
	concurrency::copy(amp_node3, node3.begin());

	p = 100;
	cout << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
	cout << p << "%\t Avg. MLUP/s " <<mlups; 
	t = (unsigned long long)param.nx*(unsigned long long)param.ny*(unsigned long long)param.t_max;
	cout<<"\nUpdated " << t << " lattice-nodes";
	diff = ( clock() - start ) / (float)CLOCKS_PER_SEC;
	cout << " in "<< diff << " seconds." << endl;

	cout << endl << "Total density: \t" << total_density3(param,node1, node2, node3) << endl;
	cout << "Avrage velocity: \t" << av_velocity3(param,node1, node2, node3,obstacle)<< endl;
	cout << "Reynolds: \t" << calc_reynolds3(param,node1, node2, node3,obstacle)<< endl;

	return 1;
}