#include "write_file.h"
#include <sstream>
#include <iomanip>

using namespace std;

int write_file3(const struct st_parameters param, std::vector<struct grid_cell1> &node1, std::vector<struct grid_cell2> &node2, std::vector<struct grid_cell3> &node3, const std::vector<int> &obstacle, int n, std::string filename)
{
	FILE* fp;						// File pointer 
	int x,y;
	int i;
	float c_sq = 1.0/3.0;			// Square of speed of sound
	float local_density = 0.0;		// Sum of densities per node
	float pressure;				// Fluid pressure in node[i]
	float u_x;						// x-component of velocity
	float u_y;						// y-component of velocity

	vector<struct write_node> writenode;
	writenode.resize(param.nx*param.ny);

	for(y=0;y<param.ny;y++) {
		for(x=0;x<param.nx;x++) {
			float density = 0;
			// Blocked node
			i = y*param.nx + x;
			if(obstacle[i] == 1){
				u_x = u_y = 0.0;
				pressure = param.density * c_sq;
			}
			// Fluid node
			else {
				local_density = 0.0;
				local_density += node3[i].n;
				local_density += node2[i].e;
				local_density += node1[i].s;
				local_density += node2[i].w;
				local_density += node3[i].ne;
				local_density += node1[i].se;
				local_density += node1[i].sw;
				local_density += node3[i].nw;
				local_density += node2[i].c;



				// Calculate x velocity component
				u_x = (node2[i].e + node3[i].ne + node1[i].se - 
						(node2[i].w + node3[i].nw + node1[i].sw)) 
						/ local_density;
				// Calculate y velocity component
				u_y = (node3[i].n + node3[i].ne + node3[i].nw - 
						(node1[i].s + node1[i].sw + node1[i].se)) 
						/ local_density;

				// Calculate pressure
				pressure = local_density * c_sq;
				density = local_density;
			}
			// Write to memory
			writenode[i].pressure = pressure;
			writenode[i].u_x = u_x;
			writenode[i].u_y = u_y;
			writenode[i].x = x;
			writenode[i].y = y;
			writenode[i].density = density;
		}
	}

	ostringstream os;

	os << "anim/" << filename << setfill('0') << setw(4) << n << ".vtk";

	fp = fopen(os.str().c_str(),"w");
	if (fp == NULL) {
		cout << "Could not open outputfile" << endl;
		return 0;
	}

	fprintf(fp, "# vtk DataFile Version 3.0 \n");
	fprintf(fp, "stream\n");
	fprintf(fp, "ASCII\n");
	fprintf(fp, "DATASET RECTILINEAR_GRID\n");
	fprintf(fp, "DIMENSIONS %d %d 1\n",param.nx,param.ny);
	fprintf(fp, "X_COORDINATES %d float\n",param.nx);

	for(x=0;x<param.nx;x++)
		fprintf(fp, "%d ",x);
	
	fprintf(fp, "\n");
	fprintf(fp, "Y_COORDINATES %d float\n",param.ny);
	
	for(y=0;y<param.ny;y++)
		fprintf(fp, "%d ",y);
	
	fprintf(fp, "\n");
	fprintf(fp, "Z_COORDINATES 1 float\n");
	fprintf(fp, "0\n");
	fprintf(fp, "POINT_DATA %d\n",param.nx*param.ny);

	fprintf(fp, "VECTORS VecVelocity float\n");

	for(y=0;y<param.ny;y++) {
		for(x=0;x<param.nx;x++) {
			i = y*param.nx + x;
				float zcord = 0;
				fprintf(fp, "%.12f %.12f %.12f\n",writenode[i].u_x,writenode[i].u_y,zcord);
		}
	}

	fprintf(fp, "SCALARS DensityDifference float 1\n");
	fprintf(fp, "LOOKUP_TABLE default\n");

	for(y=0;y<param.ny;y++) {
		for(x=0;x<param.nx;x++) {
			i = y*param.nx + x;
			fprintf(fp, "%.12f\n", writenode[i].density);
		}
	}

	fclose(fp);
	return 1;
}